/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * ftftp.cpp - the tftp fetching module header with coordinator and slaves implementation
 *
 * 2005-02-13	 you'll all die.
 * 2005-06-01	worked a little on reassembling issues
 */

#include "ftftp.h"
#include <time.h>
#include <netinet/in.h>

// uncomment the following line if you are having problems with tftp and want to debug it a little
#define VERBOSE_TFTP

using namespace mwcollectd;

extern "C"
{
  Module * __attribute__ ((dllexport)) CreateInstance(MalwareCollector * pCollector)
  {
    return ((Module *) (new TftpCoordinator(pCollector)));
  }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void TftpCoordinator::start()
{
	m_pCollector->getFetchDispatcher()->addProtocolHandler(this);

	{
		char * szPath;

		asprintf(&szPath, "%s:timeout", m_szConfigPath);
		m_ulTimeout =  (unsigned long) m_pConfig->getLong(szPath, 30);
		free(szPath);

		asprintf(&szPath, "%s:retries", m_szConfigPath);
		m_ulRetries =  (unsigned long) m_pConfig->getLong(szPath, 16);
		free(szPath);

		asprintf(&szPath, "%s:block-strict", m_szConfigPath);
		m_bBlockStrict = m_pConfig->getLong(szPath, 0) != 0;
		free(szPath);
	}
}

void TftpCoordinator::loop()
{
	for(std::list<TftpFetchSession *>::iterator i = m_lpSessions.begin(); i != m_lpSessions.end(); ++i)
		(* i)->checkTimeout();
}

void TftpCoordinator::stop()
{
	m_pCollector->getFetchDispatcher()->removeProtocolHandler(this);
}

bool TftpCoordinator::supportProtocol(const char * szProtocol)
{
	return (strcmp(szProtocol, "tftp") == 0);
}

void TftpCoordinator::fetchFile(const char * szURL)
{
	char szFile[256], szHost[256];

	{
		const char * szWalk = strstr(szURL, "://");
		unsigned int i;

		if(!szWalk)
			return;

		szWalk += 3;

		for(i = 0; * szWalk && * szWalk != '/'; ++i, ++szWalk)
			szHost[i] = * szWalk;

		szHost[i] = 0;
		++szWalk;

		strncpy(szFile, szWalk, sizeof(szFile));
	}

	{
		TftpFetchSession * pSession = new TftpFetchSession(this, m_pCollector, m_pLogManager, m_ulTimeout, m_bBlockStrict, m_ulRetries); 
		int iSocket = m_pCollector->getNetworkInterface()->createUnboundDatagramSocket(TftpCoordinator::incomingNotifier, TftpCoordinator::terminationNotifier, pSession);

		if(iSocket < 0)
			m_pLogManager->log(LL_CRITICAL, "Could not create UDP socket to fetch %s!", szURL);
		else
			pSession->fetchFile(iSocket, m_pCollector->getNetworkInterface()->lookupHost(szHost), szFile);

		m_lpSessions.push_back(pSession);
	}
}


void TftpCoordinator::terminationNotifier(bool bServer, void * pUserData)
{
	((TftpFetchSession *) pUserData)->removeMe();

	delete ((TftpFetchSession *) pUserData);
}

void TftpCoordinator::incomingNotifier(int iHandle, unsigned char * pucData, unsigned int nLength, void * pUserData, unsigned long ulAddress, unsigned short usPort)
{
	((TftpFetchSession *) pUserData)->incomingData(iHandle, pucData, nLength, ulAddress, usPort);
}

void TftpCoordinator::removeChild(TftpFetchSession * pSession)
{
	for(std::list<TftpFetchSession *>::iterator i = m_lpSessions.begin(); i != m_lpSessions.end(); ++i)
	{
		if (pSession == (*i))
		{
			m_lpSessions.erase(i);
			
			return;
		}
	}	
}



// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TftpFetchSession::~TftpFetchSession()
{

	if(m_szFileName)
		free(m_szFileName);

	if(m_pucBuffer)
		free(m_pucBuffer);

	if(m_pucLastSend)
		free(m_pucLastSend);
}

void TftpFetchSession::removeMe()
{
	m_pFather->removeChild(this);
}

void TftpFetchSession::fetchFile(int iHandle, unsigned long ulHost, const char * szFileName)
{
	{
		m_iHandle = iHandle;
		m_szFileName = strdup(szFileName);
		m_ulAddress = ulHost;
	}

	{		
		unsigned char * pucRequest = (unsigned char *) malloc(strlen(szFileName) + 9);
		m_pucLastSend = (unsigned char *) malloc(strlen(szFileName) + 9);

		* ((unsigned short *) pucRequest) = 0x0100; // read request
		strcpy((char *) pucRequest + 2, szFileName);
		strcpy((char *) pucRequest + strlen(szFileName) + 3, "octet");

		m_pCollector->getNetworkInterface()->sendDatagram(iHandle, pucRequest, strlen(szFileName) + 9, ulHost, 69);
		
		#ifdef VERBOSE_TFTP
		m_pLogManager->log(LL_SPAM, "Sent \"octet\" read request for \"%s\".", szFileName);
		#endif

		m_ulNumTimeout = 0;
		m_ulLastEvent = time(0);
		m_ulLastSend = strlen(szFileName) + 9;			
		memcpy(m_pucLastSend, pucRequest, m_ulLastSend);
		m_usPort = 69;

		free(pucRequest);
	}

	m_ulLastEvent = time(0);
	m_ulReceived = 0;
}

void TftpFetchSession::checkTimeout()
{
	if(time(0) - m_ulLastEvent >= m_ulTimeout)
	{
		if(m_ulNumTimeout < m_ulMaxTimeout)
		{
			 m_pCollector->getNetworkInterface()->sendDatagram(m_iHandle, m_pucLastSend, m_ulLastSend, m_ulAddress, m_usPort);

			 m_ulLastEvent = time(0);
		}
		else
		{
			m_pLogManager->log(LL_INFO, "Receive of %s timed out awaiting next block (%u/%u tries).", m_szFileName, m_ulNumTimeout, m_ulMaxTimeout);

			m_pCollector->getNetworkInterface()->closeSocket(m_iHandle);
		}

		++ m_ulNumTimeout;
	}
}

void TftpFetchSession::incomingData(int iHandle, unsigned char * pucData, unsigned int nLength, unsigned long ulSource, unsigned short usSource)
{
	unsigned char pucAck[4];
	unsigned short usBlock;

	if(ulSource != m_ulAddress)
		return;

	if(nLength < 4)
		return;

	if(* ((unsigned short *) pucData) != 0x0300) // no data packet
	{
		if(* ((unsigned short *) pucData) == 0x0500)
		{
			unsigned short usCode = ntohs(* ((unsigned short *) (pucData + 2)));

			if(nLength > 4 && !pucData[nLength - 1])
				m_pLogManager->log(LL_INFO, "Error fetching %s, server reported error code %u: \"%s\".", m_szFileName, usCode, pucData + 4);				
			else
				m_pLogManager->log(LL_CRITICAL, "Error fetching %s, server reported error code %u (unterminated error message)!");

			m_pCollector->getNetworkInterface()->closeSocket(iHandle);
		}

		return;
	}

	usBlock = ntohs(* ((unsigned short *) (pucData + 2)));

	m_ulLastEvent = time(0);
	
	#ifdef VERBOSE_TFTP
	m_pLogManager->log(LL_SPAM, "Got TFTP data packet, block #%u, data size %u.", (unsigned int) usBlock, nLength - 4);
	#endif
	
	{
		* ((unsigned short *) pucAck) = 0x0400;
		* ((unsigned short *) (pucAck + 2)) = htons(usBlock);
		m_pCollector->getNetworkInterface()->sendDatagram(iHandle, pucAck, 4, ulSource, usSource);
		
		#ifdef VERBOSE_TFTP
		m_pLogManager->log(LL_SPAM, "Sent TFTP acknowledgement for block #%u.", (unsigned int) usBlock);
		#endif

		memcpy(m_pucLastSend, pucAck, 4); // buffer is still filename + 9 bytes long, should be enough (+ 9 [!!!]) :D
		m_ulLastSend = 4;
		m_usPort = usSource;
		
		if(nLength > 4) // length of 0 indicated the end of a file which's size is a multiple of 512 bytes
		{
			tftp_packet_t pckPacket;
			
			m_ulReceived += nLength - 4;
			m_ulNumTimeout = 0;
			
			memcpy(pckPacket.pucData, pucData, MIN(nLength, 516));
			pckPacket.usLength = (unsigned short) nLength;
			m_lpktCache.push_back(pckPacket);
		}

		if(nLength != 516) // last packet
		{
			sortPacketsAndSubmit();

			m_pCollector->getNetworkInterface()->closeSocket(iHandle);
		}
	}
}

typedef enum
{
	SNT_CONSTANT = 0,
	SNT_INCNETBYTE,
	SNT_INCHOSTBYTE,
} sequencenumber_type_t;

void TftpFetchSession::sortPacketsAndSubmit()
{
	unsigned char * pucData;
	unsigned int nSize = 0;
	unsigned int nPointer = 0;
	sequencenumber_type_t sntType;
		
	{ // determine sequencenumber (ack number) type
		unsigned int nConstant = 0, nNetByte = 0, nHostByte = 0, nRandom = 0;
		std::list<tftp_packet_t>::iterator j;		
		
		for(std::list<tftp_packet_t>::iterator i = (++(m_lpktCache.begin())), j = m_lpktCache.begin(); i != m_lpktCache.end(); ++i, ++j)
		{
			if(* (unsigned short *) ((* i).pucData + 2) == * (unsigned short *) ((* j).pucData + 2))
				++nConstant;
			else if(* (unsigned short *) ((* i).pucData + 2) == (* (unsigned short *) ((* j).pucData + 2)) + 1)
				++nHostByte;
			else if(ntohs(* (unsigned short *) ((* i).pucData + 2)) == ntohs(* (unsigned short *) ((* j).pucData + 2)) + 1)
				++nNetByte;
			else
				++nRandom;
		}
		
		if(nRandom > nNetByte && nRandom > nHostByte && nRandom > nConstant)
			sntType = SNT_CONSTANT; // we have to guess here since seems to be purely random, sorting them as they came in should probably be the best here
		else if(nHostByte >= nNetByte && nHostByte >= nConstant)
			sntType = SNT_INCHOSTBYTE; // the bot author forgot some htons(..)
		else if(nNetByte >= nHostByte && nNetByte >= nConstant)
			sntType = SNT_INCNETBYTE; // as described in the rfc
		else if(nConstant >= nHostByte && nConstant >= nNetByte)
			sntType = SNT_CONSTANT; // simply no sequence numbers
	}
	
	if(sntType == SNT_CONSTANT)
	{ // straight forward :)
		// iterate through all packets and add the lengths to the total length
		for(std::list<tftp_packet_t>::iterator i = m_lpktCache.begin(); i != m_lpktCache.end(); ++i)
			nSize += (* i).usLength - 4;
	}
	else
	{ // might look odd at first glance, but makes sense if you look at reassembly code below
		// avoid DoS by consuming all memory by sending carefully crafted packet sequences with correct sequence numbers but one which eats all the memory
		// the + 32 is to be somewhat tolerant due to semi-good reassembly
		unsigned long ulMaximumSize = (m_lpktCache.size() + 32) * 512; 
	
		for(std::list<tftp_packet_t>::iterator i = m_lpktCache.begin(); i != m_lpktCache.end(); ++i)
		{ // iterate through all packets and if we find a new maximum size (new last packet), we take it
			unsigned short usBlock, usLength;
			
			// just determine block number and length depending on the information we learned above (htons present or not)
			if(sntType == SNT_INCHOSTBYTE)		
			{
				usBlock = * (unsigned short *) ((* i).pucData + 2);
				usLength = (* i).usLength - 4;
			}
			else
			{
				usBlock = ntohs(* (unsigned short *) ((* i).pucData + 2));
				usLength = (* i).usLength - 4;
			}
			
			// calculate the total length by taking the block number, multiply it by the data length of all packets before
			// and add the packets own size. according to rfc 1350, all packets except for the last have to hold 512 bytes of data
			if((unsigned int) (usBlock * 512 + usLength) > nSize)
				nSize = usBlock * 512 + usLength; // ok, new maximum size (new last packet), store the size
		}
		
		if(nSize > ulMaximumSize)		
		{
			m_pLogManager->log(LL_CRITICAL, "Could not sort TFTP packets by sequence number (%s byte order) due to maximum size limit.", (sntType == SNT_INCHOSTBYTE ? "host" : "network"));
			
			return;
		}
	}
	
	if(!(pucData = (unsigned char *) malloc(nSize)))
		return;
		
	// set all bytes to NOP because there might be missing packets
	memset(pucData, 0x90, nSize);
		
	if(sntType == SNT_CONSTANT)
	{
		// just add them as they came in
			
		for(std::list<tftp_packet_t>::iterator i = m_lpktCache.begin(); i != m_lpktCache.end(); ++i)
		{
			memcpy(pucData + nPointer, (* i).pucData + 4, (* i).usLength - 4);
			nPointer += (* i).usLength - 4;
		}
		
		m_pLogManager->log(LL_SPAM, "Did not sort TFTP packets due to SNT_CONSTANT.");
	}
	else if(sntType == SNT_INCHOSTBYTE || sntType == SNT_INCNETBYTE)
	{
		for(std::list<tftp_packet_t>::iterator i = m_lpktCache.begin(); i != m_lpktCache.end(); ++i)
		{
			unsigned short usBlock;
			
			// this can be easily done this way since every packet except for the last
			// has to be 512 bytes long
			// duplicate packets simple overwrite existing packets
			// missing packets are not recognized yet (TODO)
			
			if(sntType == SNT_INCHOSTBYTE)
				usBlock = * (unsigned short *) ((* i).pucData + 2);
			else
				usBlock = ntohs(* (unsigned short *) ((* i).pucData + 2));

			if(!usBlock)
			{
				m_pLogManager->log(LL_CRITICAL, "TFTP reassembly failed due to sequence number of 0!");
				free(pucData);

				return;
			}
				
			memcpy(pucData + ((usBlock - 1) * 512), (* i).pucData + 4, (* i).usLength - 4);
		}
	
		m_pLogManager->log(LL_SPAM, "Sorted TFTP packets by sequence number (%s byte order).", (sntType == SNT_INCHOSTBYTE ? "host" : "network"));
	}
	
	m_pCollector->getSubmissionDispatcher()->submitData(pucData, nSize);		
	free(pucData);
}
