/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * log-syslog.cpp - log to syslog :>
 *
 * 2005-05-29	 for grgomez ;)
 */
 
#include "log-syslog.h"
#include <stdlib.h>
#include <string.h>
#include <errno.h>

SyslogModule::SyslogModule()
{
}

SyslogModule::~SyslogModule()
{
	free(m_szConfigPath); // no checking, relies on proper utilization by core
}


void SyslogModule::setLogManager(LogManager * p)
{
	m_pLogManager = p;
}

void SyslogModule::setConfiguration(Configuration * pConfig, const char * szPath)
{
	m_pConfiguration = pConfig;
	m_szConfigPath = strdup(szPath);
}


void SyslogModule::start()
{
	const char * szIdentity;
	char * szPath;
	
	asprintf(&szPath, "%s:identity", m_szConfigPath);
	szIdentity = m_pConfiguration->getString(szPath, "mwcollectd");
	free(szPath);
	
	asprintf(&szPath, "%s:level", m_szConfigPath);
	m_llLevel = LogManager::mapLogLevel(m_pConfiguration->getString(szPath, "critical"));
	free(szPath);
	
	openlog(szIdentity, LOG_NDELAY, LOG_DAEMON);
	
	m_pLogManager->addLogger(this);
}

void SyslogModule::stop()
{
	m_pLogManager->removeLogger(this);
	
	closelog();
}


void SyslogModule::log(log_level_t llLevel, const char * szMessage)
{
	int iImportance;
	
	if(llLevel < m_llLevel)
		return;
		
	if(llLevel == LL_CRITICAL)
		iImportance = LOG_CRIT;
	else if(llLevel == LL_INFO)
		iImportance = LOG_NOTICE;
	else if(llLevel == LL_DEBUG)
		iImportance = LOG_INFO;
	else
		iImportance = LOG_DEBUG;
		
	syslog(iImportance | LOG_DAEMON, "%s", szMessage); // hah, no format string vulnerability \o/
}



extern "C"
{
  mwcollectd::Module * __attribute__ ((dllexport)) CreateInstance(MalwareCollector * pCollector)
  {
    return (new SyslogModule());
  }
}
