#include <mysql/mysql.h> 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "base64.h"
#include "db.h"

int sha1(unsigned char* data, int data_len, char* hash, int buf_len);

static MYSQL mysql;
static int INIT = 0;

static int _fetch_id (const char* hash);
static int _update_last_seen (int id);
static int _do_insert (const char* hash, const char* encdata);

int db_init (const char* host, const char* db, const char* user,
	 const char* password, unsigned int port)
{
	if (INIT == 0)
	{
		// connect to the db
		mysql_init(&mysql);
  		if (mysql_real_connect(&mysql, host, user, password, db, port,
		  NULL, 0) == NULL)
		{ 
   			printf(mysql_error(&mysql));
			return -1;
  		}

		INIT = 1;
	}

	return INIT;
}

int db_insert_file ( const char* filename )
{
  	struct stat fs;
  	FILE* fp = NULL;
  	char* data = NULL;
	int len = 0;
	int rc = 0;

	// validate input
	if (filename == NULL) 
	{
		return 0;
	}

	// get file size and open file
	if (stat(filename, &fs) == 0)
	{
		fp = fopen(filename, "rb");
	}

	// exit if open failed
	if (fp == NULL)
	{
		printf("unable to open %s\n", filename);
		exit(1);
	}

	// malloc buffer
	data = (char *) malloc(fs.st_size);
	if (data == NULL)
	{
		printf("malloc failed\n");
		exit(1);
	}	

	// read the data
	len = fread(data, 1, fs.st_size, fp);
	if (len == fs.st_size)
	{
		rc = db_insert_data((unsigned char *) data, len);
	}
	else
	{
		printf("failed to read all data from file %d != %d\n", len, 
			fs.st_size);
	}

	if (data != NULL)
	{
		free(data);
	}

	return rc;
}

int db_insert_data(const unsigned char* data, int len)
{
	char 	hash[41];
	char* 	encdata;
	int		rc = 0;

	// validate input
	if (data == NULL || len < 1)
	{
		return 0;
	}

	sha1((unsigned char *) data, len, hash, 40);
	encdata = encode_base64(len, (unsigned char*)data);
	if (encdata)
	{
		int id = _fetch_id(hash);

		if (id > 0)
		{
			rc = _update_last_seen(id);
		}
		else
		{
			rc = _do_insert(hash, encdata);
		}

		free(encdata);
	}

	return rc;
}

static int _fetch_id (const char* hash)
{
	char buffer[255];
	int rc = 0;

	if (hash == NULL || INIT != 1)
	{
		return 0;
	}

	snprintf(buffer, 254, 
					"select mw_data_id from mw_data where hash = '%s'", hash);

  	if (mysql_real_query(&mysql, buffer, strlen(buffer)) == 0)
	{
		MYSQL_RES *res;

		res = mysql_store_result(&mysql); 
		if (res)
		{
			if (mysql_num_fields(res))
			{
  				MYSQL_ROW row;
			
				if (row = mysql_fetch_row(res))
				{
					rc= atoi((char*)row[0]);
				}
			}
  			mysql_free_result(res);
		}
	}
	else
	{
   			printf(mysql_error(&mysql));
	}

	return rc;
}

static int _update_last_seen ( int id )
{
	char buffer[255];

	if (buffer == NULL || INIT != 1)
	{
		return 0;
	}

	snprintf(buffer, 254, 
		"update mw_data set last_seen_date = concat(utc_date(), ' ', utc_time()) where mw_data_id = %d", id);

  	return mysql_real_query(&mysql, buffer, strlen(buffer)) == 0 ? 1 : 0;
}

static int _do_insert (const char* hash, const char* encdata)
{
	char*	buffer = NULL;
	int		buflen = 0;
	int 	rc = 0;

	if (encdata == NULL || INIT != 1)
	{
		return 0;
	}

	buflen = strlen(encdata) + 255;
	buffer = (char *) malloc(buflen);

	if (buffer == NULL)
	{
		printf("failed to realloc buffer of %d bytes\n", buflen);
		return 0;
	}

	snprintf(buffer, buflen, "insert into mw_data (create_date, hash, data) values (concat(utc_date(), ' ', utc_time()), '%s', '%s')", hash, encdata);

  	if (mysql_real_query(&mysql, buffer, strlen(buffer)) == 0)
	{
 		rc = mysql_insert_id(&mysql); 
	}

	free(buffer);

	return rc;
}

#ifdef TEST_MAIN
int main(int argc, char **argv)
{
	int rc;

	db_init("localhost", "mwcollect", "mwcollect", "mwcollect", 0);
	rc = db_insert_file(argv[1]);
	printf("rc = %d\n", rc);
	return 0;
}
#endif

