/*
 * SubmitMySQLModule - store data in a MySQL database
 *
 */
 
#include <mwcollectd.h>

#include <stdio.h>
#include <unistd.h>

extern "C" {
#include "db.h"
};

using namespace mwcollectd;

// ============================================================

class SubmitMySQLModule : public Module, public SubmissionModule
{
public:
	SubmitMySQLModule(MalwareCollector * pCollector) { m_pCollector = pCollector; }
	virtual ~SubmitMySQLModule() { }

	virtual void setLogManager(LogManager * pLogManager) { m_pLogManager = pLogManager; }
	virtual void setConfiguration(Configuration * pConfig, const char * szPath) { m_pConfig = pConfig; m_szConfigPath = strdup(szPath); }

	virtual void start();
	virtual void stop();

	virtual module_type_t moduleType() { return MT_SUBMISSION_AGENT; }

	virtual void submitFile(const char * szFilePath);
	virtual void submitData(const unsigned char * szData, unsigned int nLength);

protected:
	MalwareCollector * m_pCollector;
	LogManager * m_pLogManager;

	Configuration * m_pConfig;
	char * m_szConfigPath;
};

// ============================================================

extern "C"
{
	Module * __attribute__ ((dllexport)) CreateInstance(MalwareCollector * pCollector)
	{
		return (Module *) (new SubmitMySQLModule(pCollector));
	}
}

// ============================================================

void SubmitMySQLModule::start()
{
	char* szPath;
	char* host;
	char* user;
	char* password;
	char* db;
	char* zport;
	unsigned int port = 0;

		asprintf(&szPath, "%s:host", m_szConfigPath);
        host = strdup(m_pConfig->getString(szPath, "localhost"));
		free(szPath);

		asprintf(&szPath, "%s:user", m_szConfigPath);
        user = strdup(m_pConfig->getString(szPath, "mwcollect"));
		free(szPath);

		asprintf(&szPath, "%s:password", m_szConfigPath);
        password = strdup(m_pConfig->getString(szPath, "mwcollect"));
		free(szPath);

		asprintf(&szPath, "%s:db", m_szConfigPath);
        db = strdup(m_pConfig->getString(szPath, "mwcollect"));
		free(szPath);

		asprintf(&szPath, "%s:port", m_szConfigPath);
        zport = strdup(m_pConfig->getString(szPath, "0"));
		free(szPath);

	port = atoi(zport);
	db_init(host, db, user, password, port);
	free(host);
	free(user);
	free(password);
	free(db);
	free(zport);

	m_pCollector->getSubmissionDispatcher()->addSubmissionModule(this);
}

void SubmitMySQLModule::stop()
{
	m_pCollector->getSubmissionDispatcher()->removeSubmissionModule(this);

	free(m_szConfigPath);
}

void SubmitMySQLModule::submitFile(const char * szFilePath)
{
	db_insert_file(szFilePath);
}

void SubmitMySQLModule::submitData(const unsigned char * szData, unsigned int nLength)
{
	db_insert_data(szData, nLength);
}
