/* mwcollect2 (c) 2004-2005 Honeynet Project
 * this code is licensed under the BSD License, the the file LICENSE for details
 *
 * shell.cpp - the shell emulation (implementation)
 *
 */

#include <mwcollectd.h>
#include <pcre.h>

namespace mwcollectd
{
	void ShellEmulation::start()
	{
		addBuiltin();
	}

	void ShellEmulation::loop()
	{
	}

	void ShellEmulation::stop()
	{
	}


	ShellSession * ShellEmulation::connectSession(unsigned long ulAddress, unsigned short usRemotePort)
	{
		if(!m_pConfig->getLong(":shell-emulation:enable-connectback", 0))
			return 0;

		ShellSession * pSession = new ShellSession(ulAddress, true, this, m_pCollector);
		int iSocket = m_pCollector->getNetworkInterface()->connectSocket(ulAddress, usRemotePort, ShellEmulation::connectedNotifier, ShellEmulation::incomingNotifier, ShellEmulation::terminationNotifier, pSession);

		pSession->m_iSocket = iSocket;
		pSession->sendHeader();

		return pSession;
	}

	ShellSession * ShellEmulation::listenSession(unsigned short usPort)
	{
		if(!m_pConfig->getLong(":shell-emulation:enable-listenshell", 0))
			return 0;		

		ShellSession * pSession = new ShellSession(0, true, this, m_pCollector);
		int iSocket = m_pCollector->getNetworkInterface()->createSocket(usPort, ShellEmulation::connectionNotifier, ShellEmulation::incomingNotifier, ShellEmulation::terminationNotifier, pSession);

		if(iSocket > 0)
		{
			m_pCollector->getNetworkInterface()->setKeepAlive(iSocket, false);

			m_pLogManager->log(LL_DEBUG, "Created listening shell session on :%u.", usPort);

			return pSession;
		}
		else
			return 0;
	}

	ShellSession * ShellEmulation::createSession(unsigned long ul)
	{
		return new ShellSession(ul, false, this, m_pCollector);
	}

	void ShellEmulation::destroySession(ShellSession * pSession)
	{
		delete pSession;
	}

	void ShellEmulation::executeCommand(const char * szCommandLine, ShellSession * pSession, unsigned int nNestLevel)
	{
		std::list<char *>::iterator j;
		std::list<ShellCommandHandler *>::iterator i;
		bool bFound = false;
		
		if(nNestLevel > 7)
		{
			m_pLogManager->log(LL_CRITICAL, "Possible exploitation condition: iNestLevel > 7 in " __FILE__ ":%i!", __LINE__);
			
			return;
		}

		if(strchr(szCommandLine, '&'))
		{
			char * szCopy = strdup(szCommandLine), * szWalk = szCopy;

            szWalk = strtok(szWalk, "&");
            executeCommand(szWalk, pSession);

            while((szWalk = strtok(0, "&")))
                executeCommand(szWalk, pSession);

			free(szCopy);
		}
		else
		{
			{
				for(i = m_lpHandlers.begin(), j =  m_lszTriggers.begin(); i != m_lpHandlers.end() && j != m_lszTriggers.end(); ++i, ++j)
				{
					if(strncmp(* j, szCommandLine, strlen(* j)) == 0)
					{
						(* i)->perform(szCommandLine, pSession);
						bFound = true;
					}
				}
			}
			
			if(!bFound && strstr(szCommandLine, ".bat"))
			{				
				if(pSession->getFile(szCommandLine) && pSession->getFile(szCommandLine)->fileType() == FT_TEXT)
				{
					std::string sLine;
					
					for(const char * c = pSession->getFile(szCommandLine)->getContent(); * c; ++c)
					{
						if(* c == '\r' || * c == '\n')
						{
							if(sLine.size())
							{
								executeCommand(sLine.c_str(), pSession, nNestLevel + 1);
								sLine.clear();
							}
						}
						else
							sLine.push_back(* c);
					}
				
					bFound = true;
					m_pLogManager->log(LL_SPAM, "Invoked batch file \"%s\".", szCommandLine);
				}
				else
					m_pLogManager->log(LL_INFO, "Did not find \"%s\" as textfile in the VFS.", szCommandLine);
			}
			else if(!bFound && strstr(".exe", szCommandLine) && pSession->getFile(szCommandLine) && pSession->getFile(szCommandLine)->fileType() == FT_EXECUTABLE)
			{
				bFound = true;
			}

			if(!bFound)
				m_pLogManager->log(LL_INFO, "Unknown Shell Command: \"%s\"", szCommandLine);
		}
	}


	void ShellEmulation::registerCommand(char * szTrigger, ShellCommandHandler * pHandler)
	{
		m_lszTriggers.push_back(szTrigger);
		m_lpHandlers.push_back(pHandler);
	}


	void ShellEmulation::addBuiltin()
	{
		ShellCommandHandler * pBasicHandler = new BasicCommandHandler(m_pLogManager);

		if(m_pConfig->getLong(":shell-emulation:tftp:enable", 0))
			registerCommand("tftp", new TftpCommandHandler(m_pCollector, m_pLogManager));

		registerCommand("echo", pBasicHandler);
		registerCommand("ftp", pBasicHandler);
        registerCommand("exit", pBasicHandler);
        registerCommand("del", pBasicHandler);
	}



	void * ShellEmulation::connectionNotifier(int iHandle, unsigned long ulSourceAddress, unsigned short usSourcePort, unsigned short usLocalPort, void * pUserData)
	{
		ShellSession * pSession = new ShellSession((ShellSession *) pUserData);

		pSession->m_iSocket = iHandle;
		pSession->setRemoteAddress(ulSourceAddress);
		pSession->sendHeader();

		return pSession;
	}

	void ShellEmulation::connectedNotifier(int iHandle, unsigned short usDestPort, void * pUserData)
	{
	}

	void ShellEmulation::incomingNotifier(int iHandle, unsigned char * pucPacketData, unsigned int nPacketLength, void * pParam)
	{
		ShellSession * pSession = (ShellSession *) pParam;

		for(unsigned int i = 0; i < nPacketLength; ++i)
		{
			if(pucPacketData[i] != '\n')
				pSession->m_sCommandBuffer.push_back(pucPacketData[i]);
			else
			{
				if(pSession->m_sCommandBuffer[pSession->m_sCommandBuffer.length() - 1] == '\r')
					pSession->m_sCommandBuffer[pSession->m_sCommandBuffer.length() - 1] = '\0';

				pSession->processBuffer();
			}
		}
	}

	void ShellEmulation::terminationNotifier(bool bServer, void * pParam)
	{
		ShellSession * pSession = (ShellSession *) pParam;

		delete pSession;
	}


	// Session Utility Functions

	void ShellSession::sendHeader()
	{
		m_pCollector->getNetworkInterface()->sendData(m_iSocket, (unsigned char *) WIN32_HEADER, sizeof(WIN32_HEADER) - 1);
	}
	
	ShellSession::~ShellSession()
	{
		for(std::list<VirtualShellFile *>::iterator i = m_lpvfFileSystem.begin(); i != m_lpvfFileSystem.end(); ++i)
			delete * i;
	}
	
	VirtualShellFile * ShellSession::addFile(const char * szFileName, filetype_t ftType)
	{
		VirtualShellFile * pFile = getFile(szFileName);
		
		if(pFile)
			return pFile;
			
		pFile = new VirtualShellFile(szFileName, ftType);
		
		m_lpvfFileSystem.push_back(pFile);
		return pFile;
	}
	
	VirtualShellFile * ShellSession::getFile(const char * szFileName)
	{
		for(std::list<VirtualShellFile *>::iterator i = m_lpvfFileSystem.begin(); i != m_lpvfFileSystem.end(); ++i)
		{
			bool fMatch = true;
			
			for(const char * c = (* i)->fileName(), * d = szFileName; * c && * d && * c != ' ' && * d != ' ' && fMatch; ++c, ++d)
				fMatch = (* c == * d);
				
			if(fMatch)
				return * i;
		}
		
		return 0;
	}


	// the builtin slaves

	TftpCommandHandler::TftpCommandHandler(MalwareCollector * pCollector, LogManager * pLogManager)
	{
		m_pCollector = pCollector;
		m_pLogManager = pLogManager;

		{
			const char * szError;
			int iPos;

			m_pPattern = pcre_compile("tftp(.exe)? +(-i +)?(.*) +get +(.*)", PCRE_CASELESS, &szError, &iPos, 0);
		}
	}

	void TftpCommandHandler::perform(const char * szCommandLine, ShellSession * pSession)
	{		
		int piOutput[5 * 3];
		int iResult;
		char * szURL;

		if(!m_pPattern)
			return;

		if((iResult = pcre_exec(m_pPattern, 0, szCommandLine, strlen(szCommandLine), 0, 0, piOutput, 5 * 3)) > 0)
		{
			const char * szAddress, * szFileName;

			pcre_get_substring(szCommandLine, piOutput, iResult, iResult - 2, &szAddress);
			pcre_get_substring(szCommandLine, piOutput, iResult, iResult - 1, &szFileName);

			if(!strcmp(szAddress, "0.0.0.0"))
			{
				struct in_addr theinaddr;

				theinaddr.s_addr = pSession->getRemoteAddress();
				asprintf(&szURL, "tftp://%s/%s", inet_ntoa(theinaddr), szFileName);
			}
			else
				asprintf(&szURL, "tftp://%s/%s", szAddress, szFileName);

			pcre_free_substring(szFileName);
			pcre_free_substring(szAddress);

			m_pCollector->getFetchDispatcher()->fetchFile(szURL);
			free(szURL);
		}
	}



	char * BasicCommandHandler::parseFtpInstructions(const char * szBuffer)
	{
		const char * szError;
		int iErrorPos;
		pcre * pPatterns[2];
		int pnOffsets[3 * 6];
		int iResult;
		
		pPatterns[0] = pcre_compile(".*open\\s+(.+)\\s+(\\d+)\\s+user\\s+(.+)\\s+(.+)\\s+.*\\s+get\\s+(\\S+)\\s+.*", PCRE_CASELESS | PCRE_UNGREEDY, &szError, &iErrorPos, 0);
		pPatterns[1] = pcre_compile(".*open\\s+(.+)\\s+(\\d+)\\s*\n\\s*(.+)\\s*\n\\s*(.+)\\s+.*\\s+get\\s+(\\S+)\\s+.*", PCRE_CASELESS | PCRE_UNGREEDY, &szError, &iErrorPos, 0);
		
		if(!pPatterns[0] || !pPatterns[1])
		{
			printf("Failed to compile pattern at #%i: %s!\n", iErrorPos, szError);
			
			return 0;
		}
		
		if((iResult = pcre_exec(pPatterns[0], 0, szBuffer, strlen(szBuffer), 0, PCRE_NOTEMPTY, pnOffsets, 3 * 6)) > 0 || (iResult = pcre_exec(pPatterns[1], 0, szBuffer, strlen(szBuffer), 0, PCRE_NOTEMPTY, pnOffsets, 3 * 6)) > 0)
		{			
			const char * szHost, * szPort, * szUser, * szPassword, * szFileName;
			char * szResult;
			
			pcre_get_substring(szBuffer, pnOffsets, iResult, 1, &szHost);
			pcre_get_substring(szBuffer, pnOffsets, iResult, 2, &szPort);
			pcre_get_substring(szBuffer, pnOffsets, iResult, 3, &szUser);
			pcre_get_substring(szBuffer, pnOffsets, iResult, 4, &szPassword);
			pcre_get_substring(szBuffer, pnOffsets, iResult, 5, &szFileName);
			
			asprintf(&szResult, "ftp://%s:%s@%s:%s/%s", szUser, szPassword, szHost, szPort, szFileName);
			
			pcre_free_substring(szFileName);
			pcre_free_substring(szPassword);
			pcre_free_substring(szUser);
			pcre_free_substring(szPort);
			pcre_free_substring(szHost);
			
			return szResult;
		}
		else
			return 0;
	}
	
	void BasicCommandHandler::perform(const char * szCommandLine, ShellSession * pSession)
	{
		if(strncmp(szCommandLine, "ftp ", 4) == 0 || strncmp(szCommandLine, "ftp.exe ", 8) == 0)
		{
			char * szInstructionFile = strstr(szCommandLine, "-s:"), * szWork;
			
			if(!szInstructionFile)
				return;
				
			szInstructionFile += 3;
			szWork = szInstructionFile;
				
			while(* szWork && * szWork != ' ')
				++szWork;
				
			* szWork = 0;
				
			if(strlen(szInstructionFile))
			{
				if(pSession->getFile(szInstructionFile))
				{
					char * szUrl = parseFtpInstructions(pSession->getFile(szInstructionFile)->getContent());
					
					if(szUrl)
					{
						const char * szFileName = szUrl + strlen(szUrl);
						
						while(* szFileName != '/')
							--szFileName;
							
						++szFileName;
						
						m_pLogManager->log(LL_SPAM, "FTP Parser: fetch \"%s\" (\"%s\")!", szUrl, szFileName);
						
						pSession->getCollector()->getFetchDispatcher()->fetchFile(szUrl);
						pSession->addFile(szFileName, FT_EXECUTABLE);
																	
						free(szUrl);
					}
				}					
				else
					m_pLogManager->log(LL_DEBUG, "FTP for non-existent \"%s\"! :(", szInstructionFile);
			}
		}
		if(strncmp(szCommandLine, "echo ", 5) == 0)
		{
			bool bAppend = false;
			char * szSave = strdup(szCommandLine + 5);
			char * szWork = szSave;
			char * szFileName = strchr(szWork, '>');
			
			if(!szFileName)
			{
				m_pLogManager->log(LL_DEBUG, "Bogus echo in shell: \"%s\"!", szCommandLine);
				return;
			}
			
			* (szFileName++) = 0;

			if(* szFileName == '>')
			{
				bAppend = true;
				++szFileName;
			}
			
			{ // trim filename
				char * szIterate;
			
				while(* szFileName == ' ')
					++szFileName;
				
				szIterate = szFileName;
			
				while(* szIterate && * szIterate != ' ')
					++szIterate;
				
				* szIterate = 0;
			}

            m_pLogManager->log(LL_SPAM, "Append \"%s\" to %s.", szWork, szFileName);
            
            VirtualShellFile * pFile = pSession->addFile(szFileName, FT_TEXT);
            
            if(bAppend)
            	pFile->appendText(szWork);
            else
            	pFile->setContent(szWork);
            	
            pFile->appendText("\n");

			free(szSave);
		}
	}
}
